#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

template <typename scalar_t>
__global__ void tiled_matmul_kernel(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias,
    scalar_t* __restrict__ output,
    int batch_size,
    int in_features,
    int out_features,
    float subtract_val,
    float multiply_val) {
    
    const int TILE = 16;
    __shared__ scalar_t input_tile[TILE][TILE];
    __shared__ scalar_t weight_tile[TILE][TILE];
    
    int row = blockIdx.x * TILE + threadIdx.x;
    int col = blockIdx.y * TILE + threadIdx.y;
    scalar_t sum = 0;

    for (int t = 0; t < (in_features + TILE - 1) / TILE; ++t) {
        int input_col = t * TILE + threadIdx.y;
        if (row < batch_size && input_col < in_features)
            input_tile[threadIdx.x][threadIdx.y] = input[row * in_features + input_col];
        else
            input_tile[threadIdx.x][threadIdx.y] = 0;

        int weight_row = col;
        int weight_col = t * TILE + threadIdx.x;
        if (weight_row < out_features && weight_col < in_features)
            weight_tile[threadIdx.x][threadIdx.y] = weight[weight_row * in_features + weight_col];
        else
            weight_tile[threadIdx.x][threadIdx.y] = 0;

        __syncthreads();

        for (int k = 0; k < TILE; ++k)
            sum += input_tile[threadIdx.x][k] * weight_tile[k][threadIdx.y];

        __syncthreads();
    }

    if (row < batch_size && col < out_features) {
        sum += bias[col];
        sum = (sum - subtract_val) * multiply_val;
        output[row * out_features + col] = max(sum, scalar_t(0));
    }
}

torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias,
    float subtract_value,
    float multiply_value) {
    
    auto batch_size = input.size(0);
    auto in_features = input.size(1);
    auto out_features = weight.size(0);
    
    auto output = torch::empty({batch_size, out_features}, input.options());
    
    dim3 blocks(
        (batch_size + 15) / 16,
        (out_features + 15) / 16
    );
    dim3 threads(16, 16);
    
    AT_DISPATCH_FLOATING_TYPES(input.type(), "tiled_matmul_kernel", ([&] {
        tiled_matmul_kernel<scalar_t><<<blocks, threads>>>(
            input.data_ptr<scalar_t>(),
            weight.data_ptr<scalar_t>(),
            bias.data_ptr<scalar_t>(),
            output.data_ptr<scalar_t>(),
            batch_size,
            in_features,
            out_features,
            subtract_value,
            multiply_value
        );
    }));
    
    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Tiled matmul with shared memory optimization");
}